# Copyright © 2019 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from toolchain.pants.auth.plugin import toolchain_auth_plugin
from toolchain.pants.auth.rules import get_auth_rules
from toolchain.pants.buildsense.rules import rules_buildsense_reporter

# Allows pants to automatically detect the remote auth plugin and activate it.
remote_auth = toolchain_auth_plugin


def rules():
    return (
        *get_auth_rules(),
        *rules_buildsense_reporter(),
    )



# Copyright © 2019 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

import datetime

UNIX_EPOCH = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
_date_fmt = "%Y/%m/%d"
_time_fmt = "%H:%M:%S"
_ms_fmt = "%f"


def utcnow() -> datetime.datetime:
    """Return a timezone aware datetime.datetime representing the current moment, in UTC.

    This is based on django.utils.timezone.now(), but without dragging django as a dependency to places that need this.
    """
    return datetime.datetime.now(tz=datetime.timezone.utc)


def datetime_fmt_std(dt, abbrev=False, ms=False):
    if dt is None:
        return None
    if dt == UNIX_EPOCH:
        return "-"
    dt = dt.astimezone(datetime.timezone.utc)
    if abbrev and dt.date() == utcnow().date():
        fmt = _time_fmt
    else:
        fmt = f"{_date_fmt} {_time_fmt}"
    if ms:
        fmt += f".{_ms_fmt}"
        return dt.strftime(fmt)[:-3]
    return dt.strftime(fmt)


def seconds_from_now(dt):
    now = utcnow()
    if dt is None or dt <= now:
        return 0
    else:
        return (dt - now).seconds



# Copyright © 2017 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# The version of the toolchain pants plugin: https://pypi.org/project/toolchain.pants.plugin/
VERSION = "0.26.0"



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# This pylint ignore is due to the migration of the pants options API, when we remove backward compatibility we should also remove this line
# pylint: disable=unexpected-keyword-arg
from __future__ import annotations

import logging
import os
import re
from typing import Mapping

from pants.option.option_types import BoolOption, IntOption, StrListOption, StrOption
from pants.option.subsystem import Subsystem

logger = logging.getLogger(__name__)


def optional_dir_option(dn: str) -> str:
    # Similar to Pant's dir_option, but doesn't require the directory to exist.
    return os.path.normpath(dn)


class CaptureCIEnv:
    _DEFAULT_CI_MAP = {
        "CIRCLECI": r"^CIRCLE.*",
        "GITHUB_ACTIONS": r"^GITHUB.*",
        # Unlike other CI systems, BITBUCKET doesn't have a default env variable that
        # indicates the it is a bitbucket environment.
        # See: https://support.atlassian.com/bitbucket-cloud/docs/variables-and-secrets/
        "BITBUCKET_BUILD_NUMBER": r"^BITBUCKET.*",
        "BUILDKITE": r"^BUILDKITE.*",
    }
    DEFAULT_EXCLUDE_TERMS = ("ACCESS", "TOKEN", "SECRET", "JWT")

    def __init__(self, *, pattern: str | None, exclude_terms: list[str], ci_map: dict[str, str] | None = None) -> None:
        self._pattern = re.compile(pattern) if pattern else None
        self._exclude_expression = re.compile("|".join(f".*{re.escape(v)}.*" for v in exclude_terms))
        self._ci_map = {ci: re.compile(pattern) for ci, pattern in (ci_map or self._DEFAULT_CI_MAP).items()}

    def _get_pattern(self, env: Mapping[str, str]) -> re.Pattern | None:
        for ci_name, capture_expression in self._ci_map.items():
            if ci_name in env:
                return capture_expression
        return None

    def capture(self, env: Mapping[str, str]) -> dict[str, str] | None:
        captured = self._capture_ci_env(env)
        return self._scrub(captured) if captured else None

    def _capture_ci_env(self, env: Mapping[str, str]) -> dict[str, str] | None:
        pattern = self._pattern or self._get_pattern(env)
        if not pattern:
            return None
        return {key: value for key, value in env.items() if pattern.match(key)}

    def _scrub(self, captured: dict[str, str]) -> dict[str, str]:
        scrubbed_keys = set()
        final_data = {}
        for key, value in captured.items():
            if self._exclude_expression.match(key):
                scrubbed_keys.add(key)
            else:
                final_data[key] = value
        captured_str = ",".join(sorted(captured.keys()))
        if scrubbed_keys:
            final_captured_str = ",".join(sorted(final_data.keys()))
            logger.debug(f"captured CI env: {captured_str} scrubbed: {scrubbed_keys} final: {final_captured_str}")
        else:
            logger.debug(f"captured CI env: {captured_str}")
        return final_data


class Reporter(Subsystem):
    options_scope = "buildsense"
    help = """Configuration for Toolchain's BuildSense reporting."""

    timeout = IntOption(
        "--timeout", advanced=True, default=10, help="Wait at most this many seconds for network calls to complete."
    )
    dry_run = BoolOption("--dry-run", advanced=True, default=False, help="Go thru the motions w/o making network calls")
    local_build_store = BoolOption(
        "--local-build-store", advanced=True, default=True, help="Store failed uploads and try to upload later."
    )
    local_store_base = StrOption(
        "--local-store-base",
        advanced=True,
        default=".pants.d/toolchain/buildsense/",
        help="Base directory for storing buildsense data locally.",
    )
    max_batch_size_mb = IntOption(
        "--max-batch-size-mb",
        advanced=True,
        default=20,
        help="Maximum batch size to try and upload (uncompressed).",
    )
    ci_env_var_pattern = StrOption(
        "--ci-env-var-pattern",
        advanced=True,
        default=None,
        help="CI Environment variables regex pattern.",
    )
    enable = BoolOption("--enable", default=True, help="Enables the BuildSense reporter plugin.")
    log_upload = BoolOption(
        "--log-upload",
        default=True,
        advanced=True,
        help="Upload pants logs to buildsense",
    )

    ci_env_scrub_terms = StrListOption(
        "--ci-env-scrub-terms",
        default=list(CaptureCIEnv.DEFAULT_EXCLUDE_TERMS),
        advanced=True,
        help="patterns for environment variables to exclude from uploaded CI env variables.",
    )

    show_link = BoolOption(
        "--show-link",
        default=True,
        advanced=True,
        help="Show link to the pants run in BuildSense Web UI.",
    )
    collect_platform_data = BoolOption(
        "--collect-platform-data",
        default=False,
        advanced=True,
        help="Should BuildSense collect and upload platform platform information (os version, platform architecture, python version, etc...).",
    )
    log_final_upload_latency = BoolOption(
        "--log-final-upload-latency",
        default=False,
        advanced=True,
        help="Should BuildSense log the time it took to upload data at the end of the run.",
    )
    batch_timeout = IntOption("--batch-timeout", advanced=True, default=40, help="Timeout for batch upload.")



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import datetime
import json
import logging
import re
from dataclasses import dataclass
from enum import Enum, unique
from pathlib import Path
from typing import Mapping

import requests

from toolchain.base.datetime_tools import utcnow
from toolchain.pants.auth.token import AuthToken
from toolchain.pants.common.network import get_common_request_headers
from toolchain.util.constants import REQUEST_ID_HEADER

_logger = logging.getLogger(__name__)

ACQUIRE_TOKEN_GOAL_NAME = "auth-acquire"  # nosec
TOKEN_EXPIRATION_WARNING_WINDOW = datetime.timedelta(days=10)


@dataclass
class RemoteClientConfig:
    cache_address: str
    exec_address: str | None


@dataclass
class TokenAndConfig:
    auth_token: AuthToken
    remote_client_cfg: RemoteClientConfig

    @classmethod
    def from_response_json(cls, resp_json: dict) -> TokenAndConfig:
        remote_cache_addr = resp_json["remote_cache"]["address"]
        remote_exec_addr = resp_json["remote_exec"]["address"] if "remote_exec" in resp_json else None
        return cls(
            auth_token=AuthToken.from_json_dict(resp_json["token"]),
            remote_client_cfg=RemoteClientConfig(cache_address=remote_cache_addr, exec_address=remote_exec_addr),
        )


@unique
class AuthState(Enum):
    UNKNOWN = "unknown"
    OK = "ok"  # We are able to auth
    UNAVAILABLE = "unavailable"  # local state is preventing auth (no or expired token file/env variable)
    FAILED = "failed"  # auth failed on the server side, don't retry (HTTP 400/403).
    TRANSIENT_FAILURE = "transient_failure"  # Server encountered a transient error (HTTP 503), it is ok to retry.

    @property
    def is_ok(self) -> bool:
        return self == self.OK

    @property
    def is_final(self) -> bool:
        return self in {self.OK, self.FAILED, self.UNAVAILABLE}

    @property
    def no_auth_possible(self) -> bool:
        return self.is_final and not self.is_ok


class AuthError(Exception):
    def __init__(
        self,
        message: str,
        context: str,
        request_id: str | None = None,
        should_retry: bool = False,
        server_failure=False,
        show_message: bool = False,
        log_level_name: str | None = None,
    ) -> None:
        message = f"[{context}] {message} request_id={request_id}" if request_id else f"[{context}] {message}"
        super().__init__(message)
        self._raw_message = message
        self._should_retry = should_retry
        self._server_failure = server_failure
        self._show_message = show_message
        self._log_level_name = log_level_name or "WARNING"

    def get_state(self) -> AuthState:
        if self._should_retry:
            return AuthState.TRANSIENT_FAILURE
        if self._server_failure:
            return AuthState.FAILED
        return AuthState.UNAVAILABLE

    @property
    def log_level(self) -> int:
        level = logging.getLevelName(self._log_level_name)
        return level if isinstance(level, int) else logging.WARNING

    def get_error_message(self) -> str:
        return self._raw_message if self._show_message else f"Error loading access token: {self!r}"


@dataclass
class AuthClient:
    context: str
    pants_bin_name: str
    base_url: str
    auth_file: str
    env_var: str | None
    repo_slug: str | None
    ci_env_variables: tuple[str, ...]
    restricted_token_matches: dict[str, str]

    @classmethod
    def create(
        cls,
        *,
        pants_bin_name: str,
        base_url: str,
        auth_file: str,
        env_var: str | None = None,
        repo_slug: str | None = None,
        ci_env_vars: tuple[str, ...] = tuple(),
        restricted_token_matches: dict[str, str] | None = None,
        context: str = "N/A",
    ):
        return cls(
            context=context,
            pants_bin_name=pants_bin_name,
            base_url=base_url,
            auth_file=auth_file,
            env_var=env_var,
            ci_env_variables=ci_env_vars,
            repo_slug=repo_slug,
            restricted_token_matches=restricted_token_matches or {},
        )

    @property
    def auth_file_path(self) -> Path:
        return Path(self.auth_file)

    def _check_refresh_token(self, refresh_token: AuthToken | None) -> None:
        if not refresh_token:
            return
        call_to_action = f"Run `{self.pants_bin_name} {ACQUIRE_TOKEN_GOAL_NAME}` to acquire a new token."
        if refresh_token.has_expired():
            raise AuthError(f"Access token has expired - {call_to_action}", context=self.context)
        time_until_expiration = refresh_token.expires_at - utcnow()
        if time_until_expiration < TOKEN_EXPIRATION_WARNING_WINDOW:
            _logger.warning(f"Access token will expire in {time_until_expiration.days} days. - {call_to_action}")

    def acquire_access_token(self, complete_env: Mapping[str, str]) -> TokenAndConfig:
        refresh_token = self.load_refresh_token(complete_env)
        self._check_refresh_token(refresh_token)
        if not refresh_token:
            return self._acquire_restricted_access_token(complete_env)
        headers = get_common_request_headers()
        _logger.debug(f"[{self.context}] Acquire access token")
        headers.update(refresh_token.get_headers())
        response = self._post(path="token/refresh/", headers=headers, timeout=4)
        resp_json = self._process_response(response)
        return TokenAndConfig.from_response_json(resp_json)

    def _acquire_restricted_access_token(self, complete_env: Mapping[str, str]) -> TokenAndConfig:
        if self._should_disable(complete_env):
            raise AuthError("Restricted token expression didn't match, disabling Toolchain auth.", context=self.context)
        env_vars = {key: complete_env[key] for key in self.ci_env_variables if key in complete_env}
        if not env_vars:
            raise AuthError(
                "Can't acquire restricted access token without environment variables.", context=self.context
            )
        headers = get_common_request_headers()
        json_data = {"repo_slug": self.repo_slug, "env": env_vars}
        _logger.info(f"[{self.context}] Acquire restricted access token: {json_data}")
        response = self._post(path="token/restricted/", headers=headers, timeout=8, json_data=json_data)
        resp_json = self._process_response(response)
        token_and_cfg = TokenAndConfig.from_response_json(resp_json)
        _logger.info(
            f"[{self.context}] restricted access token acquired. expires at: {token_and_cfg.auth_token.expires_at.isoformat()}"
        )
        return token_and_cfg

    def _post(self, path: str, headers: dict[str, str], timeout: int, json_data: dict | None = None):
        url = f"{self.base_url}/{path}"
        with requests.Session() as session:
            session.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
            try:
                return session.post(url, headers=headers, timeout=timeout, json=json_data)
            except requests.RequestException as error:
                raise AuthError(str(error), should_retry=True, context=self.context)

    def _process_server_messages(self, messages: list[dict]) -> None:
        for msg in messages:
            level = logging.getLevelName(msg.get("level", "INFO"))
            level_value = level if isinstance(level, int) else logging.INFO
            _logger.log(level=level_value, msg=msg["msg"])

    def _process_response(self, response) -> dict:
        if response.ok:
            resp_json = response.json()
            if "messages" in resp_json:
                self._process_server_messages(resp_json["messages"])
            return resp_json
        request_id = response.headers.get(REQUEST_ID_HEADER)
        if response.status_code == 503:
            raise AuthError(
                "Auth failed, transient server error",
                context=self.context,
                request_id=request_id,
                should_retry=True,
                server_failure=True,
            )
        content_type = response.headers.get("Content-Type", "N/A")
        is_json = content_type == "application/json"
        if not is_json:
            # Most likely an error we get from nginx (hence not json) and thus transient.
            raise AuthError(
                f"Auth failed, unknown error: HTTP status={response.status_code} content_type={content_type} url={response.url}",
                context=self.context,
                request_id=request_id,
                should_retry=True,
                server_failure=True,
            )
        resp_json = response.json()
        if response.status_code == 403 and resp_json.get("rejected") is True:
            detail = resp_json.get("detail", "Auth rejected by server")
            log_level = resp_json.get("log_level")
            raise AuthError(
                detail,
                context=self.context,
                request_id=request_id,
                server_failure=True,
                show_message=True,
                log_level_name=log_level,
            )
        # TODO build a better string/error message
        errors = resp_json.get("errors") or "N/A"
        raise AuthError(f"API Errors: {errors}", context=self.context, request_id=request_id, server_failure=True)

    def _should_disable(self, complete_env: Mapping[str, str]) -> bool:
        for env_var, expression in self.restricted_token_matches.items():
            if env_var not in complete_env:
                _logger.debug(f"[{self.context}] {env_var} not in env: {complete_env.keys()}")
                return True
            match = re.match(expression, complete_env[env_var]) is not None
            _logger.debug(
                f"[{self.context}] expression match={match} for {env_var}: expected={expression}  got={complete_env[env_var]}"
            )
            return not match
        return False

    def load_refresh_token(self, complete_env: Mapping[str, str]) -> AuthToken | None:
        if self.env_var:
            token = _load_from_env(self.context, complete_env, self.env_var)
            if token:
                return token
            if not self.repo_slug or not self.ci_env_variables:
                raise AuthError(
                    f"Access token not set in environment variable: {self.env_var}. customer_slug & ci_env_vars must be defined in order to acquire restricted access token.",
                    context=self.context,
                )
            return token

        auth_file_path = self.auth_file_path
        if auth_file_path.exists():
            return _load_from_file(self.context, self.pants_bin_name, auth_file_path)
        raise AuthError(
            f"Failed to load auth token (no default file or environment variable). Run `{self.pants_bin_name} {ACQUIRE_TOKEN_GOAL_NAME}` to set up authentication.",
            context=self.context,
        )


def _load_from_env(context: str, complete_env: Mapping[str, str], env_var_name: str) -> AuthToken | None:
    token_str = complete_env.get(env_var_name)
    if not token_str:
        _logger.warning(
            f"[{context}] Failed to load Toolchain token from env var '{env_var_name}'. Please make sure the env var is set in your environment."
        )
        return None
    token = AuthToken.from_access_token_string(token_str)
    _logger.info(
        f"[{context}] Successfully loaded Toolchain token from env var '{env_var_name}', expiration: {token.expires_at.isoformat()}."
    )
    return token


def _load_from_file(context: str, pants_bin_name: str, auth_file_path: Path) -> AuthToken:
    try:
        token_json = json.loads(auth_file_path.read_text())
    except (FileNotFoundError, ValueError) as err:
        raise AuthError(
            f"Failed to load auth token: {err!r}. Run `{pants_bin_name} {ACQUIRE_TOKEN_GOAL_NAME}` to set up authentication.",
            context=context,
        )
    # TODO: Handle TypeError (due to malformed json)
    return AuthToken.from_json_dict(token_json)



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# This pylint ignore is due to the migration of the pants options API, when we remove backward compatibility we should also remove this line
# pylint: disable=unexpected-keyword-arg
from __future__ import annotations

import base64
import logging
import multiprocessing
import os
import platform
import socket
import time
from dataclasses import asdict
from pathlib import Path
from threading import Thread
from typing import Mapping

from pants.engine.internals import native_engine  # type: ignore[attr-defined]
from pants.engine.streaming_workunit_handler import StreamingWorkunitContext, TargetInfo, WorkunitsCallback
from pants.option.option_value_container import OptionValueContainer

from toolchain.pants.auth.store import AuthStore
from toolchain.pants.buildsense.client import BuildSenseClient
from toolchain.pants.buildsense.common import RunTrackerBuildInfo, WorkUnits, WorkUnitsMap
from toolchain.pants.buildsense.state import BuildState
from toolchain.pants.buildsense.subsystems import CaptureCIEnv, Reporter

logger = logging.getLogger(__name__)


class ReporterCallback(WorkunitsCallback):
    """Configuration for Toolchain's BuildSense reporting."""

    def __init__(
        self,
        options: OptionValueContainer | Reporter,
        auth_store: AuthStore,
        env: Mapping[str, str],
        repo_name: str | None,
        org_name: str | None,
        base_url: str,
        git_worktree,
    ):
        super().__init__()
        self._env = env
        self._enabled = False
        if not options.enable:
            logger.debug("BuildSense plugin is disabled.")
            return
        if not repo_name:
            logger.warning("Couldn't determine repo name. BuildSense plugin will be disabled.")
            return
        client = BuildSenseClient.from_options(
            client_options=options, auth=auth_store, repo=repo_name, org_name=org_name, base_url=base_url
        )
        # set self._build_state *before* changing the state.
        self._build_state = BuildState(
            client,
            local_store_base_path=Path(options.local_store_base),
            max_batch_size_mb=options.max_batch_size_mb,
            local_store_enabled=options.local_build_store,
            log_final_upload_latency=options.log_final_upload_latency,
        )

        self._enabled = True
        self._git_worktree = git_worktree
        self._log_upload = options.log_upload
        self._call_count = 0
        self._reporter_thread = ReportThread(self._build_state)
        self._options = options
        self.__ci_capture: CaptureCIEnv | None = None
        self._build_link_logged = not options.show_link
        logger.debug("BuildSense Plugin enabled")

    @property
    def can_finish_async(self) -> bool:
        return True

    def __call__(  # type: ignore[override]
        self,
        *,
        completed_workunits: WorkUnits,
        started_workunits: WorkUnits,
        context: StreamingWorkunitContext,
        finished: bool = False,
        **kwargs,
    ) -> None:
        if not self._enabled:
            return
        self.handle_workunits(
            completed_workunits=completed_workunits,
            started_workunits=started_workunits,
            context=context,
            finished=finished,
        )

    def handle_workunits(
        self,
        *,
        completed_workunits: WorkUnits,
        started_workunits: WorkUnits,
        context: StreamingWorkunitContext,
        finished: bool,
    ) -> None:
        work_units_map = {wu["span_id"]: wu for wu in (started_workunits or [])}
        work_units_map.update({wu["span_id"]: wu for wu in (completed_workunits or [])})
        logger.debug(
            f"handle_workunits total={len(work_units_map)} completed={len(completed_workunits)} started={len(started_workunits)} finished={finished} calls={self._call_count}"
        )
        self._maybe_show_buildsense_link()
        self._build_state.set_context(context)
        if self._call_count == 0 and not finished:
            # If the first invocation of ReporterCallback by pants is also the last one
            # (i.e. if finished=True), then we don't send the initial report to buildsense.

            self._enqueue_initial_report(context)
        if finished:
            self._on_finish(context, self._call_count, work_units_map)
        else:
            self._build_state.queue_workunits(self._call_count, work_units_map)
        self._call_count += 1

    def _maybe_show_buildsense_link(self) -> None:
        if not self._build_link_logged and self._build_state.build_link:
            logger.info(f"View on BuildSense: {self._build_state.build_link}")
            self._build_link_logged = True

    @property
    def _ci_capture(self) -> CaptureCIEnv:
        if not self.__ci_capture:
            self.__ci_capture = CaptureCIEnv(
                pattern=self._options.ci_env_var_pattern,
                exclude_terms=list(self._options.ci_env_scrub_terms),
                ci_map=self._build_state.ci_capture_config,
            )
        return self.__ci_capture

    def _enqueue_initial_report(self, context: StreamingWorkunitContext) -> None:
        run_tracker_info = self._get_run_tracker_info(context)
        logger.debug(f"enqueue_initial_report {run_tracker_info.run_id}")
        self._build_state.queue_initial_report(run_tracker_info)

    def _on_finish(self, context: StreamingWorkunitContext, call_count: int, work_units_map: WorkUnitsMap) -> None:
        run_tracker_info = self._get_run_tracker_info(context)
        self._build_state.build_ended(run_tracker_info, call_count=call_count, work_units_map=work_units_map)
        self._reporter_thread.stop_thread()

    def _get_run_tracker_info(self, context: StreamingWorkunitContext) -> RunTrackerBuildInfo:
        ci_env = self._ci_capture.capture(self._env)
        run_tracker = context.run_tracker
        has_ended = run_tracker.has_ended()
        # Copy the RunTracker info before mutating it in _adjust_run_info_fields.
        run_info = dict(run_tracker.run_information())
        _adjust_run_info_fields(run_info, run_tracker.goals, has_ended, scm=self._git_worktree)

        build_stats = {
            "run_info": run_info,
            "recorded_options": run_tracker.get_options_to_record(),
        }

        if ci_env:
            build_stats["ci_env"] = ci_env

        if has_ended:
            if self._options.collect_platform_data:
                build_stats["platform"] = collect_platform_info()

            build_stats.update(
                {
                    "pantsd_stats": run_tracker.pantsd_scheduler_metrics,
                    "cumulative_timings": run_tracker.get_cumulative_timings(),  # type: ignore[dict-item]
                    "counter_names": list(run_tracker.counter_names),  # type: ignore[dict-item]
                }
            )
            targets_specs = _get_expanded_specs(context)
            if targets_specs:
                build_stats["targets"] = targets_specs
            counters = _get_counters(context)
            if counters:
                build_stats["counters"] = counters
            observation_histograms = _get_histograms(context)
            if observation_histograms:
                build_stats["observation_histograms"] = observation_histograms
        upload_log = all((has_ended, self._log_upload, run_tracker.run_logs_file))
        log_file = Path(run_tracker.run_logs_file) if upload_log else None
        return RunTrackerBuildInfo(has_ended=has_ended, build_stats=build_stats, log_file=log_file)


def _get_expanded_specs(context: StreamingWorkunitContext) -> dict[str, list[dict[str, str]]] | None:
    def to_targets_dicts(targets: list[TargetInfo]) -> list[dict[str, str]]:
        return [asdict(target) for target in targets]

    targets = context.get_expanded_specs().targets
    return {spec: to_targets_dicts(targets) for spec, targets in targets.items()}


def _get_counters(context: StreamingWorkunitContext) -> dict[str, int] | None:
    # TODO: The `get_metrics` method was added in Pants 2.11.x. Can remove when we no longer want
    # to support older versions.
    if not hasattr(context, "get_metrics"):
        return None
    return context.get_metrics()


def _get_histograms(context: StreamingWorkunitContext) -> dict | None:
    histograms_info = context.get_observation_histograms()
    version = histograms_info["version"]

    if version != 0:
        logger.warning(f"Cannot encode internal metrics histograms: unexpected version {version}")
        return None
    histograms = histograms_info["histograms"]
    if not histograms:
        return None
    return {
        "version": version,
        "histograms": {key: base64.b64encode(value).decode() for key, value in histograms.items()},
    }


def _adjust_run_info_fields(run_info: dict, goals: list[str], has_ended: bool, scm) -> None:
    host = socket.gethostname()
    run_info["machine"] = f"{host} [docker]" if _is_docker() else host
    if scm:
        revision = scm.commit_id
        run_info.update(revision=revision, branch=scm.branch_name or revision)
    else:
        logger.warning("Can't get git scm info")

    if "parent_build_id" in run_info:
        del run_info["parent_build_id"]

    run_info["computed_goals"] = goals
    if not has_ended:
        run_info["outcome"] = "NOT_AVAILABLE"


def _is_docker() -> bool:
    # Based on https://github.com/jaraco/jaraco.docker/blob/master/jaraco/docker.py
    # https://stackoverflow.com/a/49944991/38265
    cgroup = Path("/proc/self/cgroup")
    return Path("/.dockerenv").exists() or (cgroup.exists() and "docker" in cgroup.read_text("utf-8"))


class ReportThread:
    def __init__(self, build_state: BuildState) -> None:
        self._logging_destination = native_engine.stdio_thread_get_destination()
        self._build_state = build_state
        self._terminate = False
        self._reporter_thread = Thread(target=self._report_loop, name="buildsense-reporter", daemon=True)
        self._reporter_thread.start()

    def stop_thread(self):
        self._terminate = True
        self._reporter_thread.join()

    def _report_loop(self):
        native_engine.stdio_thread_set_destination(self._logging_destination)
        while not self._terminate:
            operation = self._build_state.send_report()
            if operation.is_sent():
                # If we send something in this call, then we don't need to sleep.
                continue
            time.sleep(2 if operation.is_error() else 0.05)
        self._build_state.send_final_report()


def collect_platform_info() -> dict[str, str | int]:
    platform_info = platform.uname()
    try:  # https://stackoverflow.com/a/28161352/38265
        mem_bytes = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
    except (ValueError, OSError) as err:
        logger.warning(f"failed to read memory size: {err!r}")
        mem_bytes = -1
    return {
        "os": platform_info.system,
        "os_release": platform_info.release,
        "processor": platform_info.processor,
        "python_version": platform.python_version(),
        "python_implementation": platform.python_implementation(),
        "architecture": platform.machine(),
        "cpu_count": multiprocessing.cpu_count(),
        "mem_bytes": mem_bytes,
    }



# Copyright © 2019 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# This pylint ignore is due to the migration of the pants options API, when we remove backward compatibility we should also remove this line
# pylint: disable=unexpected-keyword-arg

from __future__ import annotations

import logging
import os
import socket
import time
import uuid
import webbrowser
from dataclasses import dataclass
from pathlib import Path
from typing import Type, TypeVar
from urllib.parse import urlencode

import requests
from pants.engine.console import Console
from pants.engine.fs import CreateDigest, Digest, FileContent, Workspace
from pants.engine.goal import Goal, GoalSubsystem
from pants.engine.rules import Get, collect_rules, goal_rule, rule
from pants.option.global_options import GlobalOptions
from pants.option.option_types import BoolOption, IntOption

# Renamed in 2.15.x.
try:
    from pants.engine.env_vars import EnvironmentVars, EnvironmentVarsRequest
except ImportError:
    from pants.engine.environment import Environment as EnvironmentVars  # type: ignore[no-redef]
    from pants.engine.environment import EnvironmentRequest as EnvironmentVarsRequest  # type: ignore[no-redef]

from toolchain.base.datetime_tools import utcnow
from toolchain.pants.auth.client import ACQUIRE_TOKEN_GOAL_NAME, AuthClient, AuthError
from toolchain.pants.auth.server import AuthFlowHttpServer, TestPage
from toolchain.pants.auth.store import AuthStore
from toolchain.pants.auth.subsystems import AccessTokenAcquisitionGoalOptions, AuthStoreOptions, OutputType
from toolchain.pants.auth.token import AuthToken
from toolchain.pants.common.network import get_common_request_headers
from toolchain.pants.common.toolchain_setup import ToolchainSetup
from toolchain.util.constants import REQUEST_ID_HEADER

_logger = logging.getLogger(__name__)


def optional_file_option(fn: str) -> str:
    # Similar to Pant's file_option, but doesn't require the file to exist.
    return os.path.normpath(fn)


_Goal = TypeVar("_Goal", bound="Goal")


def _mark_environment_behavior(cls: Type[_Goal]) -> Type[_Goal]:
    """Conditionally mark a `Goal` subclass as being migrated to the new environments behavior.

    This decorator exists for backwards compatibility with Pants pre-2.15. Once support for 2.14 ends, remove this
    decorator, and move the `environment_behavior` settings directly to the `Goal` subclasses we are decorating.
    """
    if hasattr(cls, "environment_behavior"):
        cls.environment_behavior = Goal.EnvironmentBehavior.LOCAL_ONLY  # type: ignore
    return cls


class AuthTokenInfoGoalOptions(GoalSubsystem):
    name = "auth-token-info"
    help = "Show information about the current auth token used to access the Toolchain service"
    verbose = BoolOption("--verbose", default=False, help="Show all JWT claims for the current auth token")
    json = BoolOption("--json", default=False, help="Show JWT claims in JSON format (implies `verbose`).")


@_mark_environment_behavior
class AuthTokenInfo(Goal):
    subsystem_cls = AuthTokenInfoGoalOptions


class AuthTokenCheckGoalOptions(GoalSubsystem):
    name = "auth-token-check"
    help = "Check if the Toolchain auth token has expired or if it is about to expire"
    threshold = IntOption(
        "--threshold", default=14, help="Threshold in days to fail the check before the token expiration"
    )


@_mark_environment_behavior
class AuthTokenCheck(Goal):
    subsystem_cls = AuthTokenCheckGoalOptions


@_mark_environment_behavior
class AccessTokenAcquisition(Goal):
    subsystem_cls = AccessTokenAcquisitionGoalOptions


@dataclass(frozen=True)
class AccessTokenAcquisitionOptions:
    output: OutputType
    auth_options: AuthClient
    repo_name: str
    org_name: str | None
    local_port: int | None
    headless: bool
    test_page: TestPage
    description: str
    ask_for_impersonation: bool
    ask_for_remote_execution: bool

    @classmethod
    def from_options(
        cls,
        *,
        acquire_options: AccessTokenAcquisitionGoalOptions,
        store_options: AuthStoreOptions,
        pants_bin_name: str,
        repo_name: str,
        org_name: str | None,
        base_url: str,
    ) -> AccessTokenAcquisitionOptions:
        auth_opts = AuthClient.create(
            pants_bin_name=pants_bin_name,
            base_url=f"{base_url}/api/v1",
            auth_file=store_options.auth_file,
            context="auth-acquire",
        )
        return cls(
            local_port=acquire_options.local_port,
            repo_name=repo_name,
            org_name=org_name,
            auth_options=auth_opts,
            output=acquire_options.output,
            headless=acquire_options.headless,
            test_page=acquire_options.test_page,
            description=acquire_options.description or "N/A",
            ask_for_impersonation=acquire_options.for_ci,
            ask_for_remote_execution=acquire_options.remote_execution,
        )

    @property
    def log_only(self) -> bool:
        return self.output == OutputType.CONSOLE or self.ask_for_impersonation

    @property
    def base_url(self) -> str:
        return self.auth_options.base_url

    def get_auth_url(self, *, org: str | None, repo: str, params: dict[str, str]) -> str:
        params["repo"] = f"{org}/{repo}" if org else repo
        encoded_params = urlencode(params)
        return f"{self.base_url}/token/auth/?{encoded_params}"

    def get_token_exchange_url(self) -> str:
        return f"{self.base_url}/token/exchange/"

    @property
    def auth_file_path(self) -> Path:
        return self.auth_options.auth_file_path


@rule
async def construct_auth_store(
    auth_store_config: AuthStoreOptions,
    global_options: GlobalOptions,
    toolchain_setup: ToolchainSetup,
) -> AuthStore:
    environment = await Get(EnvironmentVars, EnvironmentVarsRequest(AuthStore.relevant_env_vars(auth_store_config)))
    return AuthStore(
        context="rule-construct-auth-store",
        options=auth_store_config,
        pants_bin_name=global_options.options.pants_bin_name,
        env=dict(environment),
        repo=toolchain_setup.safe_get_repo_name(),
        base_url=toolchain_setup.base_url,
    )


@goal_rule(desc="Acquires access token from Toolchain Web App and store it locally")
async def acquire_access_token(
    console: Console,
    workspace: Workspace,
    acquire_goal_options: AccessTokenAcquisitionGoalOptions,
    store_options: AuthStoreOptions,
    global_options: GlobalOptions,
    toolchain_setup: ToolchainSetup,
) -> AccessTokenAcquisition:
    repo_name = toolchain_setup.get_repo_name()
    acquire_options = AccessTokenAcquisitionOptions.from_options(
        pants_bin_name=global_options.options.pants_bin_name,
        acquire_options=acquire_goal_options,
        store_options=store_options,
        repo_name=repo_name,
        org_name=toolchain_setup.org_name,
        base_url=toolchain_setup.base_url,
    )
    if acquire_options.test_page != TestPage.NA:
        _test_local_server(acquire_options)
        return AccessTokenAcquisition(exit_code=0)
    try:
        auth_token = _acquire_token(console, acquire_options)
    except AuthError as error:
        console.print_stderr(str(error))
        return AccessTokenAcquisition(exit_code=-1)
    if acquire_options.log_only:
        console.print_stdout(f"Access Token is: {auth_token.access_token}")
        return AccessTokenAcquisition(exit_code=0)
    # stores token locally
    auth_file_path = acquire_options.auth_file_path
    digest = await Get(
        Digest, CreateDigest([FileContent(path=auth_file_path.name, content=auth_token.to_json_string().encode())])
    )
    workspace.write_digest(digest=digest, path_prefix=str(auth_file_path.parent))
    console.print_stdout("Access token acquired and stored.")
    return AccessTokenAcquisition(exit_code=0)


@goal_rule(desc=AuthTokenInfoGoalOptions.help)
async def show_token_info(
    console: Console,
    auth_store: AuthStore,
    options: AuthTokenInfoGoalOptions,
) -> AuthTokenInfo:
    for line in auth_store.get_token_info(verbose=options.verbose, use_json=options.json):
        console.print_stdout(line)
    return AuthTokenInfo(exit_code=0)


@goal_rule(desc=AuthTokenCheckGoalOptions.help)
async def check_auth_token(
    console: Console,
    auth_store: AuthStore,
    options: AuthTokenCheckGoalOptions,
) -> AuthTokenCheck:
    refresh_token = auth_store.load_refresh_token()
    if not refresh_token:
        console.print_stderr("Token not found")
        return AuthTokenCheck(exit_code=-1)
    time_until_expiration = refresh_token.expires_at - utcnow()
    if time_until_expiration.total_seconds() < 0:
        console.print_stderr(f"Token expired at {refresh_token.expires_at.isoformat()}")
        return AuthTokenCheck(exit_code=-1)
    if time_until_expiration.days <= options.threshold:
        console.print_stderr(
            f"Token will expire in {time_until_expiration.days} days at {refresh_token.expires_at.isoformat()}"
        )
        return AuthTokenCheck(exit_code=-1)
    console.print_stdout(
        f"Token will expire in {time_until_expiration.days} days at {refresh_token.expires_at.isoformat()}"
    )
    return AuthTokenCheck(exit_code=0)


def _acquire_token(console: Console, options: AccessTokenAcquisitionOptions) -> AuthToken:
    if options.headless or not _is_browser_available():
        return _acquire_token_headless(console, options)
    return _acquire_token_with_browser(console, options)


def _test_local_server(options: AccessTokenAcquisitionOptions):
    with AuthFlowHttpServer.create_server(port=options.local_port, expected_state=str(uuid.uuid4())) as http_server:
        http_server.start_thread()
        server_url = http_server.get_test_url(options.test_page)
        success = webbrowser.open(server_url, new=1, autoraise=True)
        if not success:
            http_server.shutdown()
            raise AuthError(
                f"Failed to open web browser. {ACQUIRE_TOKEN_GOAL_NAME} can't continue.", context="auth-acquire"
            )
        time.sleep(4)  # sleep to allow the browser to load the page from the server.


def _acquire_token_with_browser(console: Console, options: AccessTokenAcquisitionOptions) -> AuthToken:
    state = str(uuid.uuid4())
    with AuthFlowHttpServer.create_server(port=options.local_port, expected_state=state) as http_server:
        http_server.start_thread()
        callback_url = http_server.server_url
        console.print_stdout(f"Local Web Server running - callback at: {callback_url}")
        params = {"redirect_uri": callback_url, "state": state}
        auth_url = options.get_auth_url(org=options.org_name, repo=options.repo_name, params=params)
        _logger.debug(f"Open Browser at: {auth_url}")
        success = webbrowser.open(auth_url, new=1, autoraise=True)
        if not success:
            http_server.shutdown()
            raise AuthError(
                f"Failed to open web browser. {ACQUIRE_TOKEN_GOAL_NAME} can't continue.", context="auth-acquire"
            )
        token_code = http_server.wait_for_code()
        desc = _get_token_desc(options)
        return _exchage_code_for_token(console, options, token_code, description=desc)


def _exchage_code_for_token(
    console: Console, options: AccessTokenAcquisitionOptions, token_code: str, description: str
) -> AuthToken:
    # TODO: Use an engine intrinsic instead of directly going to the network.
    headers = get_common_request_headers()
    data = {"code": token_code, "desc": description}
    if options.ask_for_impersonation:
        data["allow_impersonation"] = "1"
    if options.ask_for_remote_execution:
        data["remote_execution"] = "1"
    with requests.post(options.get_token_exchange_url(), data=data, headers=headers) as response:
        if not response.ok:
            console.print_stderr(console.red(_get_error_message(response)))
            raise AuthError("Failed to acquire access token from server", context="auth-acquire")
        resp_data = response.json()
        return AuthToken.from_json_dict(resp_data)


def _acquire_token_headless(console: Console, options: AccessTokenAcquisitionOptions) -> AuthToken:
    url = options.get_auth_url(org=options.org_name, repo=options.repo_name, params={"headless": "1"})
    console.print_stdout(f"Using a web browser navigate to: {url}")
    # TODO: use console to get input from the user. https://github.com/pantsbuild/pants/issues/11398
    token_code = input("Type or paste in the token exchange code: ")
    desc = _get_token_desc(options)
    return _exchage_code_for_token(console, options, token_code, description=desc)


def _is_browser_available() -> bool:
    try:
        webbrowser.get()
    except webbrowser.Error:
        return False
    return True


def _get_error_message(response) -> str:
    error_message = None
    request_id = response.headers.get(REQUEST_ID_HEADER, "NA")
    if response.headers.get("Content-Type") == "application/json":
        error_message = response.json().get("message")

    if not error_message:
        error_message = f"Unknown error: {response.text}"
    return f"HTTP: {response.status_code}: {error_message} request={request_id}"


def get_auth_rules():
    return collect_rules()


def _get_token_desc(options: AccessTokenAcquisitionOptions) -> str:
    if options.description:
        return options.description
    default_desc = socket.gethostname()
    if options.log_only:
        default_desc += " [for CI]"
    # TODO: use console to get input from the user. https://github.com/pantsbuild/pants/issues/11398
    user_desc = input(f"Enter token description [{default_desc}]: ")
    return user_desc or default_desc



# Copyright © 2021 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import base64
import datetime
import json
from dataclasses import asdict, dataclass
from typing import cast

from toolchain.base.datetime_tools import utcnow


@dataclass(frozen=True)
class AuthToken:
    access_token: str
    expires_at: datetime.datetime
    user: str | None = None
    repo: str | None = None
    repo_id: str | None = None
    customer_id: str | None = None

    @classmethod
    def no_token(cls) -> AuthToken:
        return cls(access_token="", expires_at=datetime.datetime(2020, 1, 1, tzinfo=datetime.timezone.utc))  # nosec

    @classmethod
    def from_json_dict(cls, json_dict: dict):
        return cls(
            access_token=json_dict["access_token"],
            expires_at=datetime.datetime.fromisoformat(json_dict["expires_at"]),
            user=json_dict.get("user"),
            repo=json_dict.get("repo"),
            repo_id=json_dict.get("repo_id"),
            customer_id=json_dict.get("customer_id"),
        )

    @classmethod
    def from_access_token_string(cls, token_str: str) -> AuthToken:
        claims = cls.get_claims(token_str)
        expires_at = datetime.datetime.fromtimestamp(claims["exp"], tz=datetime.timezone.utc)
        return cls(
            access_token=token_str, expires_at=expires_at, user=claims["toolchain_user"], repo=claims["toolchain_repo"]
        )

    @classmethod
    def get_claims(cls, token_str):
        claims_segment = token_str.split(".")[1].encode()
        return json.loads(base64url_decode(claims_segment))

    def get_headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self.access_token}"}

    def has_expired(self) -> bool:
        expiration_time = cast(datetime.datetime, self.expires_at) - datetime.timedelta(
            seconds=10
        )  # Give some room for clock deviation and processing time.
        return utcnow() > expiration_time

    @property
    def has_token(self) -> bool:
        return bool(self.access_token)

    @property
    def claims(self) -> dict:
        return self.get_claims(self.access_token)

    @property
    def token_id(self) -> str:
        return self.claims["jid"]

    def to_json_string(self) -> str:
        token_dict = asdict(self)
        token_dict["expires_at"] = self.expires_at.isoformat()
        return json.dumps(token_dict)


def base64url_decode(data: bytes) -> bytes:
    # based on jose/utils.py
    rem = len(data) % 4
    if rem > 0:
        data += b"=" * (4 - rem)
    return base64.urlsafe_b64decode(data)



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import base64
import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path

from pants.util.dirutil import safe_concurrent_creation

from toolchain.base.datetime_tools import utcnow
from toolchain.pants.buildsense.common import Artifacts

_logger = logging.getLogger(__name__)


@dataclass
class BatchedData:
    path: Path
    version: str
    num_of_files: int

    def get_metadata(self) -> dict:
        return {"version": self.version, "num_of_files": self.num_of_files}

    @property
    def name(self) -> str:
        return self.path.name

    def get_batched_data(self) -> str:
        return self.path.read_text()


class LocalBuildsStore:
    _TIMESTAMP_FMT = "%Y_%m_%d_%H_%M_%S"
    BATCH_VERSION = "1"

    def __init__(self, base_path: Path, max_batch_size_mb: int, enabled: bool) -> None:
        self._enabled = enabled
        self._build_queue = base_path / "queue"
        self._upload_staging = base_path / "upload"
        self._build_queue.mkdir(parents=True, exist_ok=True)
        self._upload_staging.mkdir(parents=True, exist_ok=True)
        self._max_batch_size = 1024 * 1024 * max_batch_size_mb  # keep it in bytes

    def store_build(self, run_id: str, json_build_data: str) -> None:
        if not self._enabled:
            return
        _logger.debug(f"store_build run_id={run_id} data={len(json_build_data)}")
        self._queue_data(run_id, "build_stats", json_build_data)

    def store_artifacts(self, run_id: str, artifacts: Artifacts) -> None:
        if not self._enabled:
            return
        _logger.debug(f"store_artifacts run_id={run_id} data={len(artifacts)}")
        encoded_artifacts = {name: base64.encodebytes(data).decode() for name, data in artifacts.items()}
        data = {
            "run_id": run_id,
            "artifacts": encoded_artifacts,
        }
        self._queue_data(run_id, "artifacts", json.dumps(data))

    def _queue_data(self, run_id: str, data_type: str, json_data: str) -> None:
        fn = self._build_queue / data_type / f"{run_id}.json"
        with safe_concurrent_creation(fn.as_posix()) as tmp_filename, open(tmp_filename, "w") as fl:
            fl.write(json_data)

    def get_upload_batch(self) -> BatchedData | None:
        if not self._enabled:
            return None
        upload_batch = next(self._upload_staging.iterdir(), None)
        if upload_batch:
            num_of_files = self._get_num_of_files(upload_batch)
            return BatchedData(path=upload_batch, version=self.BATCH_VERSION, num_of_files=num_of_files)
        return self._create_batch()

    def _get_num_of_files(self, batch_file: Path) -> int:
        num_of_files = 0
        for files in json.loads(batch_file.read_bytes()).values():
            if isinstance(files, dict):
                num_of_files += len(files)
        return num_of_files

    def _create_batch(self) -> BatchedData | None:
        timestamp = utcnow().strftime(self._TIMESTAMP_FMT)
        batch, files_in_batch = self._get_batched_data()
        if not files_in_batch:
            return None

        batch_file = self._upload_staging / f"{timestamp}_{os.getpid()}.json"
        with safe_concurrent_creation(batch_file.as_posix()) as tmp_filename, open(tmp_filename, "w") as fl:
            fl.write(batch)
        for build_file in files_in_batch:
            build_file.unlink()
        return BatchedData(path=batch_file, version=self.BATCH_VERSION, num_of_files=len(files_in_batch))

    def _get_batched_data(self) -> tuple[str, list[Path]]:
        files_in_batch = []
        batch_data = {"version": self.BATCH_VERSION}
        # Doing some work to account for existing files, will remove after everyone runs this version of the plugin
        for sub_dir, key in [("", "build_stats"), ("build_stats", "build_stats")]:
            files_batch = self._get_files_for_batch(sub_dir)
            if not files_batch:
                continue
            files_in_batch.extend(files_batch)
            batch_data[key] = {build_file.stem: json.loads(build_file.read_bytes()) for build_file in files_batch}  # type: ignore[assignment]
        return json.dumps(batch_data), files_in_batch

    def delete_batch_file(self, batch_data: BatchedData) -> None:
        batch_file = batch_data.path
        if batch_file.exists():
            batch_file.unlink()

    def _get_files_for_batch(self, sub_dir: str) -> list[Path]:
        # Latest builds first.
        def sort_key(fl: Path) -> tuple[int, str]:
            last_modification = int(fl.stat().st_mtime / 60)  # Reduce accuracy. mostly to prevent flaky tests
            return last_modification, fl.stem

        queue_dir = self._build_queue
        if sub_dir:
            queue_dir = queue_dir / sub_dir
        if not queue_dir.exists():
            return []
        build_files = sorted(queue_dir.iterdir(), key=sort_key, reverse=True)
        files_to_upload = []
        total_size = 0
        for build_file in build_files:
            if not build_file.is_file():
                continue
            size = build_file.stat().st_size
            if size + total_size > self._max_batch_size:
                break
            total_size += size
            files_to_upload.append(build_file)
        return files_to_upload



# Copyright © 2019 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import http
import time
from contextlib import suppress
from enum import Enum, unique
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer  # type: ignore
from importlib import resources
from threading import Thread
from typing import cast
from urllib.parse import parse_qsl, urlparse


@unique
class TestPage(Enum):
    NA = "na"
    SUCCESS = "success"
    ERROR = "error"


class AuthServerHandler(BaseHTTPRequestHandler):
    CALLBACK_PATH = "/token-callback/"
    TEST_SUCCESS_PATH = "/test/success/"
    TEST_ERROR_PATH = "/test/error/"

    @property
    def auth_server(self) -> AuthFlowHttpServer:
        # this property is to silence mypy/make it happy.
        return cast(AuthFlowHttpServer, self.server)

    def _serve_favicon(self) -> None:
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "image/png")
        self.end_headers()
        self.wfile.write(self.auth_server.fav_icon)

    def _accept_code(self, parsed_url) -> None:
        query = dict(parse_qsl(parsed_url.query))
        auth_error = query.get("error")
        if auth_error:
            self._serve_error_page(auth_error)
            return
        code = query["code"]
        state = query["state"]
        success, error_message = self.auth_server.set_access_token_code(code, state)
        if not success:
            self._serve_error_page(error_message)
            return
        self._serve_success_page()

    def _serve_error_page(self, error_message, http_error_status: int = http.HTTPStatus.BAD_REQUEST) -> None:
        self.send_response(http_error_status)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(self.auth_server.get_error_html(error_message))

    def _serve_success_page(self) -> None:
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(self.auth_server.success_html)

    def do_GET(self) -> None:
        parsed_url = urlparse(self.path)
        if parsed_url.path == "/favicon.ico":
            self._serve_favicon()
        elif parsed_url.path == self.CALLBACK_PATH:
            self._accept_code(parsed_url)
        elif parsed_url.path == self.TEST_SUCCESS_PATH:
            self._serve_success_page()
        elif parsed_url.path == self.TEST_ERROR_PATH:
            self._serve_error_page("No soup for you! come back one year!")
        else:
            self._serve_error_page(f"Invalid path: {parsed_url.path}", http_error_status=http.HTTPStatus.NOT_FOUND)


class AuthFlowHttpServer(ThreadingHTTPServer):
    @classmethod
    def create_server(cls, *, port: int | None, expected_state: str):
        if port:
            return cls(port, expected_state)
        for curr_port in range(8000, 8100):
            with suppress(OSError):
                return cls(curr_port, expected_state)
        # TODO: Pants exception (TaskError)
        raise Exception("Failed to create web server")

    def __init__(self, port, expected_state):
        super().__init__(("localhost", port), AuthServerHandler)
        self._resources = {
            "favicon": _load_resource("favicon.png"),
            "success": _load_resource("success.html"),
            "error": _load_resource("error.html").decode(),
        }
        self._thread = Thread(target=self._server_thread, daemon=True)
        self._base_server_url = f"http://localhost:{port}"
        self._callback_url = f"{self._base_server_url}{AuthServerHandler.CALLBACK_PATH}"
        self._code = None
        self._expected_state = expected_state

    def _server_thread(self):
        self.serve_forever(poll_interval=0.2)

    def start_thread(self):
        self._thread.start()

    @property
    def server_url(self) -> str:
        return self._callback_url

    def get_test_url(self, test_page: TestPage) -> str:
        path = (
            AuthServerHandler.TEST_SUCCESS_PATH if test_page == TestPage.SUCCESS else AuthServerHandler.TEST_ERROR_PATH
        )
        return f"{self._base_server_url}{path}"

    @property
    def fav_icon(self) -> bytes:
        return self._resources["favicon"]

    @property
    def success_html(self) -> bytes:
        return self._resources["success"]

    def get_error_html(self, message) -> bytes:
        error_html = self._resources["error"]
        error_html = error_html.replace("$MESSAGE", message)
        return error_html.encode()

    def set_access_token_code(self, code: str, state: str) -> tuple[bool, str]:
        if not code:
            return False, "Missing token exchange code"
        if self._expected_state != state:
            return False, f"Unexpected state value: {state} (expected {self._expected_state})"
        self._code = code
        return True, ""

    def wait_for_code(self, timeout_sec: int = 300) -> str | None:
        # TODO: Failure scenarios
        timeout_time = time.time() + timeout_sec
        while not self._code and time.time() < timeout_time:
            time.sleep(0.1)
        self.shutdown()
        return self._code


def _load_resource(name: str) -> bytes:
    # Ugly hack to get local module name. can't figure out a pythonic way to do this.
    module_name = ".".join(__name__.split(".")[:-1])
    return resources.read_binary(module_name, name)



# Copyright © 2022 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

from enum import Enum, unique

from pants.engine.goal import GoalSubsystem
from pants.option.option_types import BoolOption, DictOption, EnumOption, IntOption, StrListOption, StrOption
from pants.option.subsystem import Subsystem

from toolchain.pants.auth.client import ACQUIRE_TOKEN_GOAL_NAME
from toolchain.pants.auth.server import TestPage

DEFAULT_AUTH_FILE = ".pants.d/toolchain_auth/auth_token.json"


@unique
class OutputType(Enum):
    FILE = "file"
    CONSOLE = "console"


class AccessTokenAcquisitionGoalOptions(GoalSubsystem):
    name = ACQUIRE_TOKEN_GOAL_NAME
    help = "Acquire access tokens for the Toolchain service."

    local_port = IntOption("--local-port", default=None, help="Local web server port")
    output = EnumOption(  # type: ignore[call-overload]
        "--output",
        enum_type=OutputType,
        default=OutputType.FILE,
        help="Output method for the access token. Outputting to the console is useful if the token needs to be "
        "provided to CI.",
    )
    headless = BoolOption("--headless", default=False, help="Don't open a browser when acquiring the access token")
    test_page = EnumOption(  # type: ignore[call-overload]
        "--test-page",
        enum_type=TestPage,
        default=TestPage.NA,
        advanced=True,
        help="Helper to test success and error pages without triggering the auth flow.",
    )
    description = StrOption("--description", default=None, help="Token description")
    for_ci = BoolOption(
        "--for-ci",
        default=False,
        help=f"Generate a token with impersonation permissions that can be used in CI. Limited to org owners/admins. "
        f"Implies `--{ACQUIRE_TOKEN_GOAL_NAME}-output={OutputType.CONSOLE.value}`.",
    )
    remote_execution = BoolOption(
        "--remote-execution",
        default=False,
        advanced=True,
        help="[Limited Beta] Asks for a token with remote execution permissions. This will only work for Toolchain customers that have this feature enabled.",
    )


class AuthStoreOptions(Subsystem):
    options_scope = "auth"
    help = "Set up for authenticating with Toolchain."

    auth_file = StrOption(
        "--auth-file",
        default=DEFAULT_AUTH_FILE,
        help="Path (relative to the repo root) at which to store and read the auth token.",
    )
    from_env_var = StrOption(
        "--from-env-var", default=None, help="Load the access token from this environment variable"
    )
    ci_env_variables = StrListOption(
        "--ci-env-variables",
        help="Environment variables in CI used to identify the build (used for restricted tokens)",
    )
    org = StrOption("--org", default=None, help="Organization slug for public repo PRs.")
    restricted_token_matches = DictOption[str](
        "--restricted-token-matches",
        default={},
        advanced=True,
        help="A mapping of environment variable name to a regex that must match that variable's value "
        "in order for the plugin to request a restricted access token.",
    )

    token_expiration_threshold = IntOption(
        "--token-expiration-threshold",
        default=30,
        advanced=True,
        help="Threshold (in minutes) for token TTL before plugin asks for a new token.",
    )



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

from pants.version import VERSION as PANTS_VERSION

from toolchain.pants.version import VERSION as TOOLCHAIN_VERSION


def get_common_request_headers() -> dict[str, str]:
    return {"User-Agent": f"pants/v{PANTS_VERSION} toolchain/v{TOOLCHAIN_VERSION}"}



# Copyright © 2021 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import logging
from typing import Mapping

from pants.option.global_options import AuthPluginResult, AuthPluginState
from pants.option.options import Options

from toolchain.base.datetime_tools import utcnow
from toolchain.pants.auth.rules import AuthStoreOptions
from toolchain.pants.auth.store import AuthStore
from toolchain.pants.common.toolchain_setup import ToolchainSetup

_PLUGIN_NAME = "Toolchain Remote Cache Auth Plugin"
_DISABLED_AUTH = AuthPluginResult(
    state=AuthPluginState.UNAVAILABLE,
    execution_headers={},
    store_headers={},
    instance_name=None,
    plugin_name="Toolchain Remote Cache Auth Plugin",
)

_logger = logging.getLogger(__name__)


def toolchain_auth_plugin(
    initial_execution_headers: dict[str, str],
    initial_store_headers: dict[str, str],
    options: Options,
    env: Mapping[str, str],
    prior_result: AuthPluginResult | None = None,
    **kwargs,
) -> AuthPluginResult:
    now = utcnow()
    if prior_result and prior_result.expiration and prior_result.is_available and now < prior_result.expiration:
        return prior_result
    store = _auth_store_from_options(options, env)
    if not store:
        return _DISABLED_AUTH
    access_token = store.get_access_token()
    config = store.get_remote_client_config()
    if not access_token.has_token:
        return _DISABLED_AUTH

    access_token_headers = access_token.get_headers()
    keys_for_access_token_headers = frozenset(access_token_headers.keys())

    overwritten_execution_header_keys = frozenset(initial_execution_headers.keys()).intersection(
        keys_for_access_token_headers
    )
    if overwritten_execution_header_keys:
        _logger.warning(
            f"The following remote execution header(s) will be overwritten by the Toolchain plugin: {', '.join(overwritten_execution_header_keys)}"
        )

    overwritten_store_header_keys = frozenset(initial_store_headers.keys()).intersection(keys_for_access_token_headers)
    if overwritten_store_header_keys:
        _logger.warning(
            f"The following remote store header(s) will be overwritten by the Toolchain plugin: {', '.join(overwritten_store_header_keys)}"
        )
    execution_address = config.exec_address if config else None
    return AuthPluginResult(
        state=AuthPluginState.OK,
        execution_headers={**initial_execution_headers, **access_token_headers},
        store_headers={**initial_store_headers, **access_token_headers},
        instance_name=access_token.customer_id,
        expiration=access_token.expires_at - store.token_expiration_threshold,
        store_address=config.cache_address if config else None,
        execution_address=execution_address,
        plugin_name=_PLUGIN_NAME,
    )


def _auth_store_from_options(options: Options, env: Mapping[str, str]) -> AuthStore | None:
    pants_bin_name = options.for_global_scope().pants_bin_name
    auth_options = options.for_scope(AuthStoreOptions.options_scope)
    setup_options = options.for_scope(ToolchainSetup.options_scope)
    repo_slug = setup_options.repo
    if not repo_slug:
        return None
    return AuthStore(
        context="auth-plugin",
        options=auth_options,
        pants_bin_name=pants_bin_name,
        env=env,
        repo=repo_slug,
        base_url=setup_options.base_url,
    )



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import datetime
import json
import logging
from typing import Mapping

from pants.option.option_value_container import OptionValueContainer

from toolchain.base.datetime_tools import utcnow
from toolchain.pants.auth.client import AuthClient, AuthError, AuthState, RemoteClientConfig
from toolchain.pants.auth.subsystems import AuthStoreOptions
from toolchain.pants.auth.token import AuthToken

_logger = logging.getLogger(__name__)


class AuthStore:
    def __init__(
        self,
        context: str,
        options: OptionValueContainer | AuthStoreOptions,
        pants_bin_name: str,
        env: Mapping[str, str],
        repo: str | None,
        base_url: str,
    ) -> None:
        repo_slug = f"{options.org}/{repo}" if options.org and repo else None
        self._access_token: AuthToken | None = None
        self._remote_client_cfg: RemoteClientConfig | None = None
        self._state = AuthState.UNKNOWN
        self._env = env
        self.token_expiration_threshold = datetime.timedelta(minutes=options.token_expiration_threshold)
        self._client = AuthClient.create(
            context=context,
            pants_bin_name=pants_bin_name,
            base_url=f"{base_url}/api/v1",
            auth_file=options.auth_file,
            env_var=options.from_env_var,
            ci_env_vars=tuple(options.ci_env_variables),
            repo_slug=repo_slug,
            restricted_token_matches=options.restricted_token_matches,
        )

    @staticmethod
    def relevant_env_vars(options: OptionValueContainer | AuthStoreOptions) -> tuple[str, ...]:
        env_vars = set(options.ci_env_variables)
        if options.from_env_var:
            env_vars.add(options.from_env_var)
        return tuple(env_vars)

    def _get_access_token(self) -> AuthToken | None:
        access_token = self._access_token
        if access_token and not access_token.has_expired():
            return access_token
        try:
            token_and_cfg = self._client.acquire_access_token(self._env)
        except AuthError as error:
            _logger.log(level=error.log_level, msg=error.get_error_message())
            self._state = error.get_state()
        else:
            self._access_token = token_and_cfg.auth_token
            self._remote_client_cfg = token_and_cfg.remote_client_cfg
            self._state = AuthState.OK

        return self._access_token

    def load_refresh_token(self) -> AuthToken | None:
        return self._client.load_refresh_token(self._env)

    def get_token_info(self, verbose: bool, use_json: bool) -> tuple[str, ...]:
        auth_token = self.load_refresh_token()
        if not auth_token:
            return ("Token not found",)
        days_to_exiration = (auth_token.expires_at.date() - utcnow().date()).days
        token_info = [
            f"token id: {auth_token.token_id} expires at: {auth_token.expires_at} (in {days_to_exiration} days)."
        ]
        if verbose or use_json:
            if use_json:
                # We only want to output json data in this case, so it can be parsed by other tools, so we remove the default token info line.
                token_info = [json.dumps(auth_token.claims)]
            else:
                token_info.extend(f"{k}={v}" for k, v in sorted(auth_token.claims.items()))
        return tuple(token_info)

    def get_remote_client_config(self) -> RemoteClientConfig | None:
        return self._remote_client_cfg

    def get_access_token(self) -> AuthToken:
        return self._get_access_token() or AuthToken.no_token()

    def get_auth_state(self) -> AuthState:
        if self._state.is_final:
            return self._state
        self._get_access_token()
        return self._state



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# This pylint ignore is due to the migration of the pants options API, when we remove backward compatibility we should also remove this line
# pylint: disable=unexpected-keyword-arg
from __future__ import annotations

import logging

from pants.option.option_types import StrOption
from pants.option.subsystem import Subsystem

_logger = logging.getLogger(__name__)


class ToolchainSetupError(Exception):
    """Raised if the toolchain settings are not properly configured."""


class ToolchainSetup(Subsystem):
    options_scope = "toolchain-setup"
    help = """Setup specific to the Toolchain codebase."""
    showed_warning = False
    repo = StrOption(
        "--repo",
        default=None,
        help="The name of this repo (typically its name in GitHub)",
    )
    org = StrOption(
        "--org",
        default=None,
        help="The organization name on your Toolchain account (typically the same as the org name in GitHub)",
    )
    _base_url = StrOption(
        "--base-url",
        default="https://app.toolchain.com",
        advanced=True,
        help="Toolchain base url",
    )

    def safe_get_repo_name(self) -> str | None:
        return self.repo or None

    @property
    def org_name(self) -> str:
        if not self.org:
            raise ToolchainSetupError(
                'Please set org = "<your org name>" in the [toolchain-setup] section in pants.toml.'
                "Set this to the organization name on your Toolchain account (typically the same as the org name in GitHub)."
            )
        return self.org

    @property
    def base_url(self) -> str:
        return self._base_url

    def get_repo_name(self) -> str:
        repo = self.safe_get_repo_name()
        if not repo:
            raise ToolchainSetupError("Repo must be set under toolchain-setup.repo.")
        return repo



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import json
import logging
import zlib
from dataclasses import dataclass
from urllib.parse import urljoin

import requests
from pants.goal.run_tracker import RunTrackerOptionEncoder
from pants.option.option_value_container import OptionValueContainer
from requests.exceptions import RequestException

from toolchain.pants.auth.store import AuthStore
from toolchain.pants.auth.token import AuthToken
from toolchain.pants.buildsense.common import Artifacts, WorkUnits
from toolchain.pants.buildsense.subsystems import Reporter
from toolchain.pants.common.network import get_common_request_headers
from toolchain.util.constants import REQUEST_ID_HEADER

_logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class BuildSenseRunInfo:
    ci_user_api_id: str | None
    # the server can decide not to return a link in order to disable this showing the link on the client side.
    build_link: str | None


@dataclass(frozen=True)
class RunEndInfo:
    success: bool
    data: str
    # the server can decide not to return a link in order to disable this showing the link on the client side.
    build_link: str | None


class BuildSenseClient:
    _IMPERSONATE_HEADER = "X-Toolchain-Impersonate"

    @classmethod
    def from_options(
        cls,
        client_options: OptionValueContainer | Reporter,
        auth: AuthStore,
        repo: str,
        org_name: str | None,
        base_url: str,
    ) -> BuildSenseClient:
        return cls(
            base_url=f"{base_url}/api/v1/repos/",
            auth=auth,
            repo=repo,
            org_name=org_name,
            timeout=client_options.timeout,
            dry_run=client_options.dry_run,
            batch_timeout=client_options.batch_timeout,
        )

    def __init__(
        self,
        *,
        base_url: str,
        repo: str,
        org_name: str | None,
        auth: AuthStore,
        timeout: int,
        batch_timeout: int,
        dry_run: bool,
    ) -> None:
        self._base_url = base_url
        self._successful_calls = 0
        self._dry_run = dry_run
        self._timeout = timeout
        self._batch_timeout = batch_timeout
        self._auth = auth
        self._repo = repo
        self._org = org_name
        self._compression_threshold = 10 * 1024  # 10kb
        self._session = requests.Session()
        self._session.mount(self._base_url, requests.adapters.HTTPAdapter(max_retries=3))
        self._last_error: type | None = None

    def _get_headers(self, user_api_id: str | None, token: AuthToken) -> dict[str, str]:
        headers = get_common_request_headers()
        headers.update(token.get_headers())
        headers["X-Pants-Stats-Version"] = "3"
        if user_api_id:
            headers[self._IMPERSONATE_HEADER] = user_api_id
        return headers

    def is_auth_available(self) -> bool:
        return not self._auth.get_auth_state().no_auth_possible

    @property
    def has_successful_calls(self) -> bool:
        # We want multiple successful calls before we return true here which will trigger a possible batch upload.
        return self._successful_calls > 2

    def _call_success(self) -> None:
        self._successful_calls += 1
        self._last_error = None

    def _prepare_request(self, api_call_desc: str, data: str) -> tuple[dict, dict, int]:
        headers = {}
        compression = len(data) > self._compression_threshold
        uncompressed_size = len(data)
        if compression:
            headers["Content-Encoding"] = "compress"
            data = zlib.compress(data.encode())  # type: ignore
            compressed_size = len(data)
            rate = compressed_size / uncompressed_size
            compression_str = f" compressed={compressed_size:,} rate={rate:.2%}"
        else:
            compression_str = ""
            compressed_size = len(data)

        _logger.debug(
            f"send_json_data {api_call_desc}. compression={compression} payload_size={uncompressed_size:,}{compression_str}"
        )
        return {"files": {"buildsense": data}}, headers, compressed_size

    def _send_json_data(
        self,
        *,
        api_call_desc: str,
        method: str,
        path: str,
        data: str,
        user_api_id: str | None,
        timeout: int | None = None,
    ) -> dict | None:
        url = urljoin(self._base_url, path)
        req_kwargs, headers, compressed_size = self._prepare_request(api_call_desc, data)
        token = self._auth.get_access_token()
        if not token.has_token:
            return None
        headers.update(self._get_headers(user_api_id, token))
        if self._dry_run:
            return None
        try:
            response = self._session.request(
                method=method, url=url, headers=headers, timeout=timeout or self._timeout, **req_kwargs  # type: ignore[arg-type]
            )
            response.raise_for_status()
        except RequestException as error:
            self._process_error(api_call_desc, compressed_size, error)
            return None
        # Only returning a response if we got a HTTP 20x
        self._call_success()
        return response.json()

    def _send_files_data(
        self,
        *,
        api_call_desc: str,
        path: str,
        files_data: dict[str, bytes],
        compressed: bool,
        user_api_id: str | None,
    ) -> bool:
        token = self._auth.get_access_token()
        url = urljoin(self._base_url, path)
        if not token.has_token:
            return False
        headers = self._get_headers(user_api_id, token)
        if compressed:
            headers["Content-Encoding"] = "compress"
        if self._dry_run:
            return True
        try:
            response = self._session.post(url=url, headers=headers, timeout=self._timeout, files=files_data)  # type: ignore[arg-type]
            response.raise_for_status()
        except RequestException as error:
            self._process_error(api_call_desc, sum(len(fl) for fl in files_data.values()), error)
            return False
        # Only returning a response if we got a HTTP 20x
        self._call_success()
        return True

    def _process_error(self, api_call_desc: str, payload_size: int, error: RequestException) -> None:
        # TODO: handle errors better.
        response = error.response
        if response is not None:  # bool(response) will return False for failed requests.
            request_id = response.headers.get(REQUEST_ID_HEADER, "N/A")
            error_message = f" request id: {request_id} error message: {error.response.text}"
        elif error.request:
            error_message = error.request.url
        else:
            error_message = ""
        level = logging.DEBUG if self._last_error == type(error) else logging.WARNING
        self._last_error = type(error)
        self._successful_calls = 0
        _logger.log(
            level=level,
            msg=f"Error reporting {api_call_desc}. payload_size={payload_size:,}: {error!r} {error_message}",
        )

    @property
    def _url_base(self) -> str:
        return f"{self._org}/{self._repo}/buildsense" if self._org else f"{self._repo}/buildsense"

    def submit_workunits(self, *, run_id: str, call_num: int, user_api_id: str | None, workunits: WorkUnits):
        wu_json = json.dumps({"workunits": workunits, "run_id": run_id})
        _logger.debug(f"workunits(call_num={call_num}, work_units={len(workunits)})")
        self._send_json_data(
            api_call_desc=f"workunits(call_num={call_num}, work_units={len(workunits)})",
            method="post",
            path=f"{self._url_base}/{run_id}/workunits/",
            data=wu_json,
            user_api_id=user_api_id,
        )

    def submit_run_start(self, *, run_id: str, build_stats: dict) -> BuildSenseRunInfo | None:
        data = json.dumps(build_stats, cls=RunTrackerOptionEncoder)
        json_response = self._send_json_data(
            api_call_desc="run_start",
            method="post",
            path=f"{self._url_base}/{run_id}/",
            data=data,
            user_api_id=None,
        )
        return (
            BuildSenseRunInfo(ci_user_api_id=json_response.get("ci_user_api_id"), build_link=json_response.get("link"))
            if json_response is not None
            else None
        )

    def submit_run_end(self, *, run_id: str, user_api_id: str | None, build_stats: dict) -> RunEndInfo:
        """Sends the final build report to the service.

        If it fails to upload, it returns the data so it can be queued for upload at a later time.
        """
        data = json.dumps(build_stats, cls=RunTrackerOptionEncoder)
        json_response = self._send_json_data(
            api_call_desc="run_end",
            method="patch",
            path=f"{self._url_base}/{run_id}/",
            data=data,
            user_api_id=user_api_id,
        )
        return (
            RunEndInfo(success=True, data=data, build_link=json_response.get("link"))
            if json_response
            else RunEndInfo(success=False, data=data, build_link=None)
        )

    def submit_batch(self, *, batched_data: str, batch_name: str, user_api_id: str | None) -> bool:
        response = self._send_json_data(
            api_call_desc=f"batch_upload={batch_name}",
            method="post",
            path=f"{self._url_base}/batch/",
            data=batched_data,
            user_api_id=user_api_id,
            timeout=self._batch_timeout,
        )
        return bool(response)

    def upload_artifacts(self, *, run_id, artifacts: Artifacts, user_api_id: str | None) -> bool:
        files_data = {name: zlib.compress(data) for name, data in artifacts.items()}
        return self._send_files_data(
            api_call_desc=f"upload_artifacts artifacts={len(artifacts)}",
            path=f"{self._url_base}/{run_id}/artifacts/",
            files_data=files_data,
            compressed=True,
            user_api_id=user_api_id,
        )

    def get_plugin_config(self) -> dict | None:
        url = urljoin(self._base_url, f"{self._url_base}/")
        token = self._auth.get_access_token()
        if not token.has_token:
            return None
        headers = get_common_request_headers()
        headers.update(token.get_headers())
        if self._dry_run:
            return None
        try:
            response = self._session.options(url=url, headers=headers, timeout=3)
            response.raise_for_status()
        except RequestException as error:
            self._process_error("get_config", payload_size=0, error=error)
            return None
        # Only returning a response if we got a HTTP 20x
        return response.json()



# Copyright © 2022 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# Must be the same as frontend/src/utils/constants.ts
REQUEST_ID_HEADER = "X-Request-Id"



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import json
import logging
import uuid
from dataclasses import dataclass
from typing import Dict, Iterator, Union, cast

import chardet
from pants.engine.fs import Digest, DigestContents, FileDigest, Snapshot

from toolchain.pants.buildsense.common import Artifacts, WorkUnits, WorkUnitsMap

_NAMES = {"start": ("start_secs", "start_nanos"), "duration": ("duration_secs", "duration_nanos")}

_logger = logging.getLogger(__name__)


Artifact = Union[str, bytes, Dict[str, str]]
ArtifactsMap = Dict[str, Artifact]

_logger = logging.getLogger(__name__)


def _to_usecs(data, name) -> int:
    secs, nanos = _NAMES[name]
    return (data[secs] * 1_000_000) + round(data[nanos] / 1_000)


@dataclass(frozen=True)
class ArtifactRef:
    wu_id: str
    name: str
    digest: Digest | None = None
    snapshot: Snapshot | None = None

    @classmethod
    def create(cls, wu_id: str, name: str, artifact: Digest | Snapshot):
        if isinstance(artifact, Digest):
            digest = artifact
            snapshot = None
        else:
            digest = None
            snapshot = artifact
        return cls(
            wu_id=wu_id,
            name=name,
            digest=digest,
            snapshot=snapshot,
        )

    @property
    def key(self) -> str:
        return f"{self.wu_id}_{self.name}"


class WorkUnitConverter:
    _ARTIFACT_SIZE_FILE_THRESHOLD = 1024 * 1024 * 30  # over 30kb
    _ALLOWED_ENCODINGS = frozenset(("ascii", "utf-8"))

    @classmethod
    def from_server(cls, config: dict, snapshot_type: type) -> WorkUnitConverter:
        # Config is defined in buildsense/ingestion/views_api.py:BuildsenseConfig
        wu_config = config["work_units"]
        _logger.debug(f"WorkUnitConverter: Config from server: {wu_config}")
        return cls(
            snapshot_type=snapshot_type,
            allowed_artifacts_keys=frozenset(wu_config["artifacts"]),
            allowed_metadata=frozenset(wu_config["metadata"]),
        )

    @classmethod
    def create_local(cls, snapshot_type: type) -> WorkUnitConverter:
        _logger.warning("WorkUnitConverter: Fallback on hard coded config")
        return cls(
            snapshot_type=snapshot_type,
            allowed_artifacts_keys=frozenset(("stdout", "stderr", "xml_results")),
            allowed_metadata=frozenset(("exit_code", "definition", "source", "address", "addresses")),
        )

    def __init__(
        self, snapshot_type: type, allowed_artifacts_keys: frozenset[str], allowed_metadata: frozenset[str]
    ) -> None:
        self._workunit_map: WorkUnitsMap = {}
        self._allowed_levels = {"INFO"}

        self._allowed_standalone_artifacts = frozenset(("coverage_xml",))
        self._allowed_metadata = allowed_metadata
        self._allowed_artifacts_keys = allowed_artifacts_keys
        self._standalone_artifacts: dict[str, ArtifactRef] = {}
        self._snapshot_type = snapshot_type

    def set_context(self, context) -> None:
        self._context = context

    def _maybe_capture_standalone_artifacts(self, wu_json: dict, name: str, artifact: Digest | Snapshot) -> bool:
        size = (
            artifact.digest.serialized_bytes_length  # type: ignore[union-attr]
            if isinstance(artifact, self._snapshot_type)
            else artifact.serialized_bytes_length  # type: ignore[union-attr]
        )
        if size < self._ARTIFACT_SIZE_FILE_THRESHOLD and name not in self._allowed_standalone_artifacts:
            return False
        ref = ArtifactRef.create(wu_json["span_id"], name, artifact)
        if ref.key not in self._standalone_artifacts:
            self._standalone_artifacts[ref.key] = ref
        return True

    def transform(self, workunits: WorkUnitsMap, call_num: int, last_update_timestamp: int) -> WorkUnits:
        self._workunit_map.update(workunits)
        return [self._wu_dict(wu, call_num, last_update_timestamp) for wu in self._filter_work_units(workunits)]

    def get_all_work_units(self, call_num: int, last_update_timestamp: int) -> WorkUnits:
        all_wus = [
            self._wu_dict_with_artifacts(wu, call_num, last_update_timestamp) for wu in self._workunit_map.values()
        ]
        if self._standalone_artifacts:
            _logger.debug(f"Collected standalone artifacts: {self._standalone_artifacts}")
        return all_wus

    def get_standalone_artifacts(self) -> Artifacts | None:
        if not self._standalone_artifacts:
            return None
        return self._get_standalone_snapshot_artifacts()

    def _get_standalone_snapshot_artifacts(self) -> Artifacts:
        refs = []
        snapshots = []
        for ref in self._standalone_artifacts.values():
            if not ref.snapshot:
                continue
            snapshots.append(ref.snapshot)
            refs.append(ref)
        artifacts_files = {}
        artifacts_descriptors = {}

        for index, digest_contents in enumerate(self._context.snapshots_to_file_contents(snapshots)):
            ref = refs[index]
            for fc in digest_contents:
                filename = uuid.uuid1().hex
                artifacts_files[filename] = fc.content
                artifacts_descriptors[filename] = {
                    "workunit_id": ref.wu_id,
                    "name": ref.name,
                    "path": fc.path,
                }
        artifacts_files["descriptors.json"] = json.dumps(artifacts_descriptors).encode()
        return artifacts_files

    def _filter_work_units(self, workunits: WorkUnitsMap) -> Iterator[dict]:
        reported_ids = set()
        for wu_id, wu_json in workunits.items():
            if wu_id in reported_ids:
                continue
            should_report = wu_json.get("artifacts") or wu_json["level"] in self._allowed_levels
            if not should_report:
                continue
            reported_ids.add(wu_id)
            parent_id = wu_json.get("parent_id")
            while parent_id:
                parent_wu = self._workunit_map[parent_id]
                wu_id = parent_wu["span_id"]
                if wu_id not in reported_ids:
                    reported_ids.add(wu_id)
                    yield parent_wu
                parent_id = parent_wu.get("parent_id")
            yield wu_json

    def _wu_dict(self, wu_json: dict, version: int, last_update_timestamp: int) -> dict:
        """Create a work unit json in a structure expected by the buildsense service."""
        start_usecs = _to_usecs(wu_json, "start")
        is_finished = "duration_secs" in wu_json
        wu_id = wu_json["span_id"]
        if "parent_ids" in wu_json:
            parent_ids = wu_json["parent_ids"]
        else:
            parent_id = wu_json.get("parent_id")
            parent_ids = [parent_id] if parent_id else []
        wu = {
            "workunit_id": wu_id,
            "name": wu_json["name"],
            "state": "finished" if is_finished else "started",
            "version": version,
            "parent_ids": parent_ids or [],
            "last_update": last_update_timestamp,
            "start_usecs": start_usecs,
        }
        description = wu_json.get("description")
        if description:
            wu["description"] = description

        if is_finished:
            duration_usecs = _to_usecs(wu_json, "duration")
            wu["end_usecs"] = start_usecs + duration_usecs
        return wu

    def _wu_dict_with_artifacts(self, wu_json: dict, version: int, last_update_timestamp: int) -> dict:
        wu = self._wu_dict(wu_json, version, last_update_timestamp)
        counters = wu_json.get("counters")
        metadata = wu_json.get("metadata", {}).items()
        allowed_metadata = {key: value for key, value in metadata if key in self._allowed_metadata}
        if counters:
            wu["counters"] = counters
        if allowed_metadata:
            wu["metadata"] = allowed_metadata
        artifacts = self._get_artifacts(wu_json)
        if not artifacts:
            return wu
        wu["artifacts"] = artifacts
        return wu

    def _get_digest_artifacts(self, wu_json: dict, keys: list[str], digests: list[FileDigest | Digest]) -> ArtifactsMap:
        if not keys:
            return {}
        try:
            artifacts = (artifact.decode() for artifact in self._context.single_file_digests_to_bytes(digests))
        except Exception:
            _logger.exception(f"Failed to get digests. keys={keys} digests={digests} {wu_json}")
            raise
        return dict(zip(keys, artifacts))

    def _get_snapshot_artifacts(self, keys: list[str], snapshots: list[Snapshot]) -> ArtifactsMap:
        if not keys:
            return {}

        def dump_dc(dc: DigestContents) -> dict[str, str]:
            # Primitive mechanism using chardet to make sure we don't try to dump binary data here.
            # if needed, binary data should be captured via standalone artifacts
            return {
                fc.path: fc.content.decode()
                for fc in dc
                if chardet.detect(fc.content)["encoding"] in self._ALLOWED_ENCODINGS
            }

        artifacts = (dump_dc(dc) for dc in self._context.snapshots_to_file_contents(snapshots))
        return dict(zip(keys, artifacts))

    def _get_artifacts(self, wu_json: dict) -> ArtifactsMap | None:
        digest_keys: list[str] = []
        digests: list[FileDigest | Digest] = []
        snapshot_keys: list[str] = []
        snapshots: list[Snapshot] = []
        for key, artifact in wu_json.get("artifacts", {}).items():
            if not isinstance(artifact, (self._snapshot_type, Digest, FileDigest)):
                _logger.warning(f"unexpected artifact type: {type(artifact)} {artifact}")
                continue
            if self._maybe_capture_standalone_artifacts(wu_json, key, artifact):  # type: ignore[arg-type]
                continue
            if isinstance(artifact, self._snapshot_type) and artifact.digest.serialized_bytes_length:  # type: ignore [attr-defined]
                snapshot_keys.append(key)
                snapshots.append(cast(Snapshot, artifact))
            elif isinstance(artifact, (Digest, FileDigest)) and artifact.serialized_bytes_length:
                digest_keys.append(key.replace("_digest", ""))
                digests.append(artifact)  # type: ignore[arg-type]
        # This is somewhat wasteful since we ask pants for artifacts we are going to end up filtering out.
        # However, this will let us know if there artifacts we can't read.
        # Eventually, we will need to better optimize it.
        artifacts = self._get_digest_artifacts(wu_json, digest_keys, digests)
        artifacts.update(self._get_snapshot_artifacts(snapshot_keys, snapshots))
        filtered_artifacts = {key: value for key, value in artifacts.items() if key in self._allowed_artifacts_keys}
        return filtered_artifacts or None




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/toolchain:toolchain-pants-plugin

from setuptools import setup

setup(**{
    'additional_classifiers': [
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    'author': 'Toolchain Inc',
    'author_email': 'info@toolchain.com',
    'classifiers': [
        'Intended Audience :: Developers',
        'License :: Other/Proprietary License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Software Development :: Build Tools',
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    'description': 'Toolchain Pants plugin.',
    'entry_points': {
        'pantsbuild.plugin': [
            'rules = toolchain.pants.register:rules',
            'remote_auth = toolchain.pants.auth.plugin:toolchain_auth_plugin',
        ],
    },
    'install_requires': (
        'chardet==5.0.0',
        'requests>=2.28.1',
    ),
    'license': 'Proprietary',
    'long_description': """# Toolchain Pants plugin

A Pants build system plugin providing support for the Toolchain platform.

Speed up your builds dramatically through shared execution and caching.
View and search your team's entire build history. Use data to find
bottlenecks and optimize your development cycle.
""",
    'long_description_content_type': 'text/markdown',
    'name': 'toolchain.pants.plugin',
    'namespace_packages': (
    ),
    'package_data': {
        'toolchain.pants.auth': (
            'error.html',
            'favicon.png',
            'success.html',
        ),
    },
    'packages': (
        'toolchain',
        'toolchain.base',
        'toolchain.pants',
        'toolchain.pants.auth',
        'toolchain.pants.buildsense',
        'toolchain.pants.common',
        'toolchain.util',
    ),
    'project_urls': {
        'Changelog': 'https://docs.toolchain.com/docs/toolchain-pants-plugin-changelog',
        'Documentation': 'https://docs.toolchain.com/docs/getting-started-with-toolchain',
        'Twitter': 'https://twitter.com/toolchainlabs',
    },
    'python_requires': '>=3.7.2,<3.10',
    'url': 'https://toolchain.com',
    'version': '0.26.0',
})



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

Artifacts = Dict[str, bytes]
WorkUnits = List[dict]
WorkUnitsMap = Dict[str, dict]


@dataclass
class RunTrackerBuildInfo:
    has_ended: bool
    build_stats: dict
    log_file: Path | None

    @property
    def run_id(self) -> str:
        return self.build_stats["run_info"]["id"]



# Copyright © 2020 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

from __future__ import annotations

import logging
import queue
import time
from enum import Enum, unique
from pathlib import Path
from typing import Tuple

from pants.engine.fs import Snapshot

from toolchain.base.datetime_tools import utcnow
from toolchain.pants.buildsense.client import BuildSenseClient
from toolchain.pants.buildsense.common import Artifacts, RunTrackerBuildInfo, WorkUnitsMap
from toolchain.pants.buildsense.converter import WorkUnitConverter
from toolchain.pants.buildsense.local_store import LocalBuildsStore

logger = logging.getLogger(__name__)
WorkUnitsChunk = Tuple[int, WorkUnitsMap, int]


class InvalidStateError(Exception):
    """Raised when there is an invalid state in the reporter logic."""


@unique
class ReportOperation(Enum):
    NO_OP = "no_op"
    SENT = "sent"
    ERROR = "error"

    def is_sent(self) -> bool:
        return self == self.SENT

    def is_error(self) -> bool:
        return self == self.ERROR


class BuildState:
    _PANTS_LOG_ARTIFACT_NAME = "pants_run_log"

    def __init__(
        self,
        client: BuildSenseClient,
        local_store_base_path: Path,
        max_batch_size_mb: int,
        local_store_enabled: bool,
        snapshot_type: type = Snapshot,
        log_final_upload_latency: bool = True,
    ):
        self._enabled = True
        self._client = client
        self._initial_report: RunTrackerBuildInfo | None = None
        self._end_report_data: tuple[RunTrackerBuildInfo, int] | None = None
        self._sent_initial_report = False
        self._run_id: str | None = None
        self._ci_user_api_id: str | None = None
        self._workunits_chunks_queue: queue.LifoQueue = queue.LifoQueue(maxsize=300)
        self._snapshot_type = snapshot_type
        self.__converter: WorkUnitConverter | None = None
        self._pants_context = None
        self._local_store = LocalBuildsStore(
            base_path=local_store_base_path, max_batch_size_mb=max_batch_size_mb, enabled=local_store_enabled
        )
        self._sumbit_batches = True
        self._build_link: str | None = None
        self._config: dict | None = None
        self._asked_for_config = False
        self._log_final_upload_latency = log_final_upload_latency

    @property
    def build_link(self) -> str | None:
        return self._build_link

    @property
    def ci_capture_config(self) -> dict[str, str] | None:
        return self._config["ci_capture"] if self._config else None

    def set_context(self, context) -> None:
        self._pants_context = context

    def _get_converter(self) -> WorkUnitConverter:
        converter = self.__converter
        if not converter:
            converter = (
                WorkUnitConverter.from_server(self._config, self._snapshot_type)
                if self._config
                else WorkUnitConverter.create_local(self._snapshot_type)
            )
        if self._pants_context:
            converter.set_context(self._pants_context)
        self.__converter = converter
        return converter

    def queue_workunits(self, call_num: int, workunits: WorkUnitsMap, timestamp: int | None = None) -> None:
        if not self._enabled:
            # Don't queue up work units when buildsense is disabled, wastes memory.
            return
        timestamp = timestamp or int(utcnow().timestamp())
        run_id = self._run_id
        if not workunits:
            return
        if not run_id:
            raise InvalidStateError("run_id must be initialized.")
        self._workunits_chunks_queue.put_nowait((call_num, workunits, timestamp))

    def set_run_id(self, run_tracker_info: RunTrackerBuildInfo) -> None:
        self._run_id = run_tracker_info.run_id

    def queue_initial_report(self, run_tracker_info: RunTrackerBuildInfo) -> None:
        self.set_run_id(run_tracker_info)
        self._initial_report = run_tracker_info

    def build_ended(self, run_tracker_info: RunTrackerBuildInfo, call_count: int, work_units_map: WorkUnitsMap) -> None:
        self.set_run_id(run_tracker_info)
        self.queue_workunits(call_count, work_units_map)
        self.submit_final_report(run_tracker_info, call_count)

    def submit_final_report(self, run_tracker_info: RunTrackerBuildInfo, calls_count: int) -> None:
        if not run_tracker_info.has_ended:
            logger.warning("RunTracker.end() was not called")
            return
        self._run_id = self._run_id or run_tracker_info.run_id
        logger.debug(
            f"submit_final_report run_id={self._run_id} calls={calls_count} keys={run_tracker_info.build_stats.keys()}"
        )
        self._end_report_data = run_tracker_info, calls_count

    def send_report(self) -> ReportOperation:
        # Expects to be non-reentrant
        if not self._enabled:
            return ReportOperation.ERROR
        if not self._client.is_auth_available():
            logger.warning("Auth failed - BuildSense plugin is disabled.")
            self._enabled = False
            return ReportOperation.ERROR
        self._maybe_get_config()
        operation = self._maybe_send_initial_report()
        if operation.is_error():
            return operation
        sent = self._send_workunit_queue()
        if sent:
            return ReportOperation.SENT
        batch_sent = self._maybe_send_batched_builds()
        return ReportOperation.SENT if batch_sent or operation.is_sent() else ReportOperation.NO_OP

    def _get_log_file(self, run_tracker_info: RunTrackerBuildInfo) -> bytes | None:
        if not run_tracker_info.log_file:
            logger.warning("no log file associated in run tracker")
            return None
        log_data = run_tracker_info.log_file.read_bytes()
        logger.debug(f"get_log_file log_file={run_tracker_info.log_file.as_posix()} size={len(log_data)}")
        return log_data

    def _get_artifacts(self, converter: WorkUnitConverter, run_tracker_info: RunTrackerBuildInfo) -> Artifacts:
        artifacts = converter.get_standalone_artifacts() or {}
        pants_run_log = self._get_log_file(run_tracker_info)
        if pants_run_log:
            artifacts[self._PANTS_LOG_ARTIFACT_NAME] = pants_run_log
        return artifacts

    def send_final_report(self) -> None:
        if self._end_report_data is None:
            raise InvalidStateError("End report data not captured.")
        start = time.time()
        run_tracker_info, calls_count = self._end_report_data
        build_stats = run_tracker_info.build_stats
        # Make sure we don't miss any work units we didn't de-queue yet
        call_num, workunits, timestamp = self._get_lastest_workunits()
        converter = self._get_converter()
        converter.transform(workunits, call_num, timestamp)
        all_workunits = converter.get_all_work_units(call_num=calls_count, last_update_timestamp=int(time.time()))
        build_stats["workunits"] = all_workunits
        run_id = self._run_id
        if not run_id:
            logger.warning("run_id is missing")
            return
        if not build_stats:
            logger.warning("final build report missing")
            return
        artifacts = self._get_artifacts(converter, run_tracker_info)
        logger.debug(f"send_final_report workunits={len(all_workunits)} artifacts={len(artifacts)}")
        run_end_info = self._client.submit_run_end(
            run_id=run_id, user_api_id=self._ci_user_api_id, build_stats=build_stats
        )
        if run_end_info.success:
            self._build_link = self._build_link or run_end_info.build_link
        else:
            # If submission failed, submit_run_end returns the data so we can store it for later upload
            self._local_store.store_build(run_id=run_id, json_build_data=run_end_info.data)
            return
        self._maybe_upload_artifacts(run_id, artifacts)
        latency = time.time() - start
        if self._log_final_upload_latency:
            logger.info(f"BuildSense final upload took {latency:.3f} seconds.")

    def _maybe_upload_artifacts(self, run_id: str, artifacts: Artifacts) -> None:
        if not artifacts:
            return
        self._client.upload_artifacts(run_id=run_id, artifacts=artifacts, user_api_id=self._ci_user_api_id)
        # TODO: Store artifacts for later in self._local_store if upload fails.

    def _maybe_get_config(self) -> None:
        if self._asked_for_config:
            return
        self._asked_for_config = True
        config_response = self._client.get_plugin_config()
        if config_response:
            self._config = config_response["config"]
        logger.debug(f"Got config from server: {(self._config or {}).keys()}")

    def _maybe_send_initial_report(self) -> ReportOperation:
        if self._sent_initial_report:
            return ReportOperation.NO_OP
        run_tracker_info = self._initial_report
        if self._sent_initial_report or not run_tracker_info:
            return ReportOperation.NO_OP
        run_info = self._client.submit_run_start(
            run_id=run_tracker_info.run_id, build_stats=dict(run_tracker_info.build_stats)
        )
        if not run_info:
            return ReportOperation.ERROR
        self._ci_user_api_id = run_info.ci_user_api_id
        self._sent_initial_report = True
        self._build_link = run_info.build_link
        return ReportOperation.SENT

    def _send_workunit_queue(self) -> bool:
        call_num, workunits, timestamp = self._get_lastest_workunits()
        if not workunits:
            return False
        logger.debug(f"send_workunit_queue workunits={len(workunits)} call={call_num}")
        run_id = self._run_id
        if not run_id:
            raise InvalidStateError("run_id must be initialized.")
        converter = self._get_converter()
        buildsense_workunits = converter.transform(workunits, call_num, timestamp)
        if not buildsense_workunits:
            return False
        self._client.submit_workunits(
            run_id=run_id, call_num=call_num, user_api_id=self._ci_user_api_id, workunits=buildsense_workunits
        )
        return True

    def _maybe_send_batched_builds(self) -> bool:
        if not self._client.has_successful_calls and self._sumbit_batches:
            return False
        batched_build = self._local_store.get_upload_batch()
        if not batched_build:
            return False
        success = self._client.submit_batch(
            batched_data=batched_build.get_batched_data(),
            batch_name=batched_build.name,
            user_api_id=self._ci_user_api_id,
        )
        if success:
            self._local_store.delete_batch_file(batched_build)
        else:
            # If we failed, don't try again during this run
            self._sumbit_batches = False
        return True

    def _empty_queue(self) -> list[WorkUnitsChunk]:
        data = []
        while self._workunits_chunks_queue.not_empty:
            try:
                data.append(self._workunits_chunks_queue.get_nowait())
            except queue.Empty:
                break
        return data

    def _get_lastest_workunits(self) -> WorkUnitsChunk:
        data = self._empty_queue()
        if not data:
            return -1, dict(), 0
        workunits_dict: WorkUnitsMap = {}
        # Ensures that we handle the latest chunks first (based on the call_num)
        # Protects against a race in which data is queued while we try to empty the queue.
        data = sorted(data, reverse=True, key=lambda x: x[0])
        for _, wu_chunk, _ in data:
            for wu_id, workunit in wu_chunk.items():
                workunits_dict.setdefault(wu_id, workunit)
        last_chunk = data[0]
        last_call_num = last_chunk[0]
        timestamp = last_chunk[2]
        return last_call_num, workunits_dict, timestamp



# Copyright © 2022 Toolchain Labs, Inc. All rights reserved.
#
# Toolchain Labs, Inc. CONFIDENTIAL
#
# This file includes unpublished proprietary source code of Toolchain Labs, Inc.
# The copyright notice above does not evidence any actual or intended publication of such source code.
# Disclosure of this source code or any related proprietary information is strictly prohibited without
# the express written permission of Toolchain Labs, Inc.

# This pylint ignore is due to the migration of the pants options API, when we remove backward compatibility we should also remove this line
# pylint: disable=unexpected-keyword-arg
from __future__ import annotations

import logging

from pants.engine.rules import Get, collect_rules, rule
from pants.engine.streaming_workunit_handler import WorkunitsCallbackFactory, WorkunitsCallbackFactoryRequest
from pants.engine.unions import UnionRule
from pants.vcs.git import GitWorktreeRequest, MaybeGitWorktree

# Renamed in 2.15.x.
try:
    from pants.engine.env_vars import CompleteEnvironmentVars
except ImportError:
    from pants.engine.environment import CompleteEnvironment as CompleteEnvironmentVars  # type: ignore[no-redef]

from toolchain.pants.auth.store import AuthStore
from toolchain.pants.buildsense.reporter import Reporter, ReporterCallback
from toolchain.pants.common.toolchain_setup import ToolchainSetup

logger = logging.getLogger(__name__)


class BuildsenseCallbackFactoryRequest:
    """A unique request type that is installed to trigger construction of our WorkunitsCallback."""


@rule
async def construct_buildsense_callback(
    _: BuildsenseCallbackFactoryRequest,
    reporter: Reporter,
    toolchain_setup: ToolchainSetup,
    auth_store: AuthStore,
    environment: CompleteEnvironmentVars,
) -> WorkunitsCallbackFactory:
    repo_name = toolchain_setup.safe_get_repo_name()
    maybe_gwt = await Get(MaybeGitWorktree, GitWorktreeRequest())
    git_worktree = maybe_gwt.git_worktree

    return WorkunitsCallbackFactory(
        lambda: ReporterCallback(
            reporter,
            auth_store=auth_store,
            env=dict(environment),
            repo_name=repo_name,
            org_name=toolchain_setup.org_name,
            base_url=toolchain_setup.base_url,
            git_worktree=git_worktree,
        )
    )


def rules_buildsense_reporter():
    return [
        UnionRule(WorkunitsCallbackFactoryRequest, BuildsenseCallbackFactoryRequest),
        *collect_rules(),
    ]
